/** @file
 * @brief Public PadicoTM module API
 * @ingroup Puk
 */

/** @example Tutorial/01-HelloWorld/HelloWorld.c
 */

#ifndef PADICO_MODULE_H
#define PADICO_MODULE_H

#include "Puk.h"

#ifdef __cplusplus
#define EXTERN_C_BEGIN extern "C" {
#define EXTERN_C_END }
#define EXTERN_C extern "C"
#else
#define EXTERN_C_BEGIN
#define EXTERN_C_END
#define EXTERN_C extern
#endif

#define PUK_MOD_INTERNALS_FUNCTIONS(MODNAME) \
  static inline padico_mod_t padico_module_self(void) \
    { return __padico_mod_instance_##MODNAME; } \
  static inline const char*padico_getattr(char*label) \
    { return puk_mod_getattr(padico_module_self(), label); } \
  static inline const char*padico_getlocalattr(char*label) \
    { return puk_mod_local_getattr(padico_module_self(), label); } \
  static inline void padico_setattr(char*label, char*value) \
    { puk_mod_setattr(padico_module_self(), label, value); } \
  static inline const char*padico_module_self_name(void) \
    { return (puk_mod_getname(padico_module_self())); } \
  static inline padico_mod_t padico_module_self_handle(void) \
    { return (padico_module_self()); }

/* #define here for a "soft link" to the binary-driver functions */
#define padico_module_myjob()  padico_tm_binary_myjob()
#define padico_module_exit(RC) padico_tm_binary_exit(RC)

#define PUK_MOD_INTERNALS_IMPORT(MODNAME) \
  EXTERN_C padico_mod_t __padico_mod_instance_##MODNAME;

#ifdef __cplusplus
#define PUK_MOD_INTERNALS_CREATE(MODNAME) \
  EXTERN_C padico_mod_t __padico_mod_instance_##MODNAME; \
           padico_mod_t __padico_mod_instance_##MODNAME = NULL; \
  EXTERN_C const char   __padico_mod_varname[] = "__padico_mod_instance_"#MODNAME; \
  EXTERN_C void padico_module_setself(padico_mod_t s) \
    { __padico_mod_instance_##MODNAME = s; } 
#else /* __cplusplus */
#define PUK_MOD_INTERNALS_CREATE(MODNAME) \
           padico_mod_t __padico_mod_instance_##MODNAME = NULL; \
           const char   __padico_mod_varname[] = "__padico_mod_instance_"#MODNAME; \
  static inline void __padico_module_setself(padico_mod_t s) { __padico_mod_instance_##MODNAME = s; } \
  EXTERN_C void padico_module_setself(padico_mod_t s) { __padico_module_setself(s); }
#endif  /* __cplusplus */

/* *** public macros */

#define PADICO_MODULE_HOOK(MODNAME) \
  PUK_MOD_INTERNALS_IMPORT(MODNAME); \
  PUK_MOD_INTERNALS_FUNCTIONS(MODNAME);

#define PADICO_MODULE_DECLARE(MODNAME) \
  PUK_MOD_INTERNALS_CREATE(MODNAME); \
  PUK_MOD_INTERNALS_FUNCTIONS(MODNAME);

#define PADICO_MODULE_INIT(INITFUNC) \
  static int INITFUNC(void); \
  EXTERN_C int padico_module_init(void) { return (INITFUNC()); }

#define PADICO_MODULE_RUN(RUNFUNC) \
  static int RUNFUNC(int argc, char**argv); \
  EXTERN_C int padico_module_run(int argc, char**argv) { return (RUNFUNC(argc, argv)); }

#define PADICO_MODULE_FINALIZE(FINALIZEFUNC) \
  static void FINALIZEFUNC(void); \
  EXTERN_C void padico_module_finalize(void) { FINALIZEFUNC(); }


#endif /* PADICO_MODULE_H */
