/** @file Launcher stub for 'LegacyBinary' service.
 */

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>

/* defined in padico_d.c */
extern void padico_d_start(int argc, char* argv[]);

/* defined in libc */
extern char**environ;

extern int  padico_launch_legacy_binary(int argc, char**argv);

static int  padico_launch_main(int argc, char**argv, char**arge);
static void padico_launch_init(void);


/* ********************************************************* */

static struct padico_launch_s
{
  int (*_main) (int, char **, char **);
  int _argc;
  char**_ubp_av;
  void (*_init) (void);
  void (*_fini) (void);
  void (*_rtld_fini) (void);
  void*_stack_end;
  char**_environ;
} padico_launch;

typedef int (*__libc_start_main_t)(int (*main) (int, char **, char **),
				   int argc,
				   char**ubp_av,
				   void (*init) (void),
				   void (*fini) (void),
				   void (*rtld_fini) (void),
				   void*stack_end);

/* ********************************************************* */

static int padico_launch_main(int argc, char**argv, char**arge)
{
  fprintf(stderr, "padico-launch: starting padico-d...\n");
  padico_launch._environ = arge;
  padico_d_start(argc, argv);
  return 0;
}

static void padico_launch_init(void)
{
  if(padico_launch._init)
    {
      (*padico_launch._init)();
    }
}

int __libc_start_main(int (*main) (int, char **, char **),
		      int argc,
		      char**ubp_av,
		      void (*init) (void),
		      void (*fini) (void),
		      void (*rtld_fini) (void),
		      void*stack_end)
{
  int rc = 0;
  fprintf(stderr, "padico-launch: intercept __libc_start_main()\n");
  padico_launch = (struct padico_launch_s)
    {
      ._main      = main,
      ._argc      = argc,
      ._ubp_av    = ubp_av,
      ._init      = init,
      ._fini      = fini,
      ._rtld_fini = rtld_fini,
      ._stack_end = stack_end
    };
#if 0
  __libc_start_main_t start_main = dlsym(RTLD_NEXT, "__libc_start_main");
  if(!start_main)
    {
      char*err = dlerror();
      fprintf(stderr, "Puk: [legacy] error %s\n", err);
      abort();
    }
  fprintf(stderr, "padico-launch: starting real __libc_start_main()...\n");
  rc = (*start_main)(&padico_launch_main,
		     padico_launch._argc,
		     padico_launch._ubp_av,
		     &padico_launch_init,
		     padico_launch._fini,
		     padico_launch._rtld_fini,
		     padico_launch._stack_end);
#else /* bourrin */
  padico_launch_init();
  padico_launch_main(padico_launch._argc, padico_launch._ubp_av, environ);
  (*padico_launch._fini)();
#endif
    
  fprintf(stderr, "padico-launch: exiting __libc_start_main()\n");
  return rc;
}

/* typeof(__libc_start_main) local__libc_start_main  __attribute__((alias("__libc_start_main"))); */

/* the real 'main' from the application is started here.
 * this function is actually called by the 'start' method of
 * the 'LegacyBinary' driver.
 */
extern int padico_launch_legacy_binary(int argc, char**argv)
{
  int rc = (*padico_launch._main)(argc, argv, environ);
  return rc;
}

