/** @file
 * PadicoTM generic bootstrap process
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "padico_config.h"
#ifdef PADICO_ENABLE_PUKABI
#include "Puk-ABI.h"
#endif

#define PADICO_PUK_LIB_NAME    DYNLIB_PREFIX "PadicoPuk" DYNLIB_EXT
#define PADICO_PUK_ENTRY_POINT SYMBOL_PREFIX "padico_puk_bootstrap"

typedef void (*main_fun_t)(int, char**);
void padico_bugfix();

static char padico_puk_lib_file[PATH_MAX];

static void init_path(void)
{
  const int len = PATH_MAX;
  const char* padico_root = getenv("PADICO_ROOT");
  if (padico_root == NULL)
    {
      fprintf(stderr,"padico-d: PADICO_ROOT environment variable is not set!\n");
      abort();
    }
  if(snprintf(padico_puk_lib_file, len, "%s/lib/%s", padico_root, PADICO_PUK_LIB_NAME) > len)
    {
      fprintf(stderr,"padico-d: error while building Puk lib name.\n");
      abort();
    }
}

extern void padico_d_start(int argc, char* argv[])
{
  void*handle = NULL;
  main_fun_t puk_boot = NULL;
  int err = -1;

  init_path();
#ifdef PADICO_ENABLE_PUKABI
  /* defeat LD_PRELOAD inheritance! */
  puk_abi_init();
#endif
  fprintf(stderr, "padico-d: loading Puk (%s)...\n", PADICO_PUK_LIB_NAME);

  handle = dlopen(padico_puk_lib_file, RTLD_NOW | RTLD_GLOBAL);
  err = errno;
  if (handle == NULL)
    {
      fprintf(stderr, "padico-d: unable to find Padico micro kernel library %s (%s).\n",
	      padico_puk_lib_file, dlerror());
      abort();
    }

  puk_boot = dlsym(handle, PADICO_PUK_ENTRY_POINT);
  
  if (puk_boot == NULL)
    {
      fprintf(stderr, "padico-d: %s symbol not found in %s (%s).\n",
	      PADICO_PUK_ENTRY_POINT,
	      PADICO_PUK_LIB_NAME, dlerror());
      abort();
    }
  (*puk_boot)(argc, argv);
     
  fprintf(stderr, "padico-d: closing Puk...\n");

  dlclose(handle);
  fprintf(stderr, "padico-d: exit.\n");

  _exit(0);
}


int main(int argc, char* argv[])
{
  padico_d_start(argc, argv);
  return 0;
}


/* dependancies for some Fortran symbols */

void getarg_()
{}

void iargc_()
{}

void pmpi_wtick_()
{}

void pmpi_wtime_()
{}

void padico_bugfix_mpi_f77()
{
  getarg_();
  iargc_();
  pmpi_wtick_();
  pmpi_wtime_();
}

void padico_bugfix()
{
  padico_bugfix_mpi_f77();
}

