/** @file
 * @brief Internal definitions for use inside Puk only.
 * @ingroup Puk
 */

/** @defgroup Puk Core module: Puk -- PadicoTM micro-kernel
 * @author Alexandre Denis
 */


#ifndef PUK_INTERNALS_H
#define PUK_INTERNALS_H

#include "Puk.h"

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

__PUK_SYM_INTERNAL void padico_puk_trace_init(void);
__PUK_SYM_INTERNAL void padico_puk_pkg_init(void);
__PUK_SYM_INTERNAL void padico_puk_lib_init(void);
__PUK_SYM_INTERNAL void padico_puk_monitor_init(void);
__PUK_SYM_INTERNAL void padico_puk_mod_init(void);
__PUK_SYM_INTERNAL void padico_puk_xml_init (void);
__PUK_SYM_INTERNAL void padico_puk_boot_init(void);
__PUK_SYM_INTERNAL void padico_puk_components_init(void);

/** @internal */
__PUK_SYM_INTERNAL void padico_puk_trace_finalize(void);


/* ** attrs */

struct padico_attr_s
{
  const char* label;
  const char* value;
};
typedef struct padico_attr_s*padico_attr_t;

/* ** units */

PUK_VECT_TYPE(padico_unit, struct padico_unit_s);

/* ** mod_refs */
struct puk_mod_ref_s
{
  char*name;
  int is_resolved;
  padico_mod_t mod;
};
typedef struct puk_mod_ref_s*puk_mod_ref_t;
PUK_VECT_TYPE(puk_mod_ref, struct puk_mod_ref_s);

/* ** module */

struct padico_mod_s
{
  char*                     mod_name;    /**< name of the module */
  puk_driver_t              driver;      /**< driver used to manage this module */
  const char*               driver_name; /**< name of the driver */
  struct padico_unit_vect_s units;       /**< units- content of the module */
  puk_hashtable_t           attrs;       /**< attributes of the module */
  padico_mod_t              inherits;    /**< ancestor for attributes inheriting */
  struct puk_mod_ref_vect_s requires;    /**< dependancies- may be unresolved until the module is actually loaded */
  int                       ref_count;   /**< number of references (other modules using this one or explicit 'load') */ 
};

__PUK_SYM_INTERNAL
puk_hashtable_enumerator_t puk_mod_enumerate_drivers(void);


#endif /* PUK_INTERNALS_H */
