/** @file
 * @brief Puk automatic lists definition
 * @ingroup Puk
 */

  /* ********* PUK_LIST ********* */

/**@hideinitializer
 * Generates a new type for list of elements */
#define PUK_LIST_TYPE(ENAME, DECL) \
  PUK_LIST_DECLARE_TYPE(ENAME) \
  PUK_LIST_CREATE_TYPE(ENAME, DECL)

/**@hideinitializer
 * Forward type declaration for lists */
#define PUK_LIST_DECLARE_TYPE(ENAME) \
  /** automatic Puk type: ENAME##_list_t is a list of ENAME##_t */ \
  typedef struct ENAME##_list_s* ENAME##_list_t; \
  /** automatic Puk type: defines ENAME##_t */ \
  typedef struct ENAME##_s* ENAME##_t; \
  /** automatic Puk type: ENAME##_itor_t is an iterator on lists of ENAME##_t */ \
  typedef ENAME##_t ENAME##_itor_t;

/**@hideinitializer
 * The effective type declaration for lists */
#define PUK_LIST_CREATE_TYPE(ENAME, DECL) \
  /** from Puk automatic type: ENAME##_t */ \
  struct ENAME##_s \
  { \
    struct ENAME##_s*_prev; /**< @internal previous cell */ \
    struct ENAME##_s*_next; /**< @internal next cell */ \
    DECL \
  }; \
  /** @internal */ \
  struct ENAME##_list_s \
  { \
    struct ENAME##_s* _head; /**< @internal head of the list */ \
    struct ENAME##_s* _tail; /**< @internal tail of the list */ \
  }; \
  /** @internal */static inline ENAME##_t ENAME##_new(void) \
    { ENAME##_t e = (ENAME##_t)padico_malloc(sizeof(struct ENAME##_s)); \
      e->_next = NULL; e->_prev = NULL; return e; } \
  /** @internal */static inline void ENAME##_delete(ENAME##_t e) \
    { padico_free(e); } \
  /** @internal */static inline void ENAME##_list_push_front(ENAME##_list_t l, ENAME##_t e) \
    { if(l->_tail == NULL) l->_tail = e; else l->_head->_prev = e; \
      e->_prev = NULL; e->_next = l->_head; l->_head = e; } \
  /** @internal */static inline void ENAME##_list_push_back(ENAME##_list_t l, ENAME##_t e) \
    { if(l->_head == NULL) l->_head = e; else l->_tail->_next = e; \
      e->_next = NULL; e->_prev = l->_tail; l->_tail = e; } \
  /** @internal */static inline ENAME##_t ENAME##_list_front(ENAME##_list_t l) \
    { return l->_head; } \
  /** @internal */static inline ENAME##_t ENAME##_list_back(ENAME##_list_t l) \
    { return l->_tail; } \
  /** @internal */static inline void ENAME##_list_init(ENAME##_list_t l) \
    { l->_head=NULL; l->_tail=l->_head; } \
  /** @internal */static inline ENAME##_list_t ENAME##_list_new(void) \
    { ENAME##_list_t l; l=(ENAME##_list_t)padico_malloc(sizeof(struct ENAME##_list_s)); \
      ENAME##_list_init(l); return l; } \
  /** @internal */static inline int ENAME##_list_empty(ENAME##_list_t l) \
    { return (l->_head == NULL); } \
  /** @internal */static inline void ENAME##_list_delete(ENAME##_list_t l) \
    { padico_free(l); } \
  /** @internal */static inline void ENAME##_list_erase(ENAME##_list_t l, ENAME##_t c) \
    { ENAME##_t p = c->_prev; if(p) p->_next = c->_next; else l->_head = c->_next; \
      if(c->_next) c->_next->_prev = p; else l->_tail = p; } \
  /** @internal */static inline ENAME##_t ENAME##_list_pop_front(ENAME##_list_t l) \
    { ENAME##_t e = ENAME##_list_front(l); \
      ENAME##_list_erase(l, e); return e; } \
  /** @internal */static inline ENAME##_itor_t ENAME##_list_begin(ENAME##_list_t l) \
    { return l->_head; } \
  /** @internal */static inline ENAME##_itor_t ENAME##_list_end(ENAME##_list_t l) \
    { return NULL; } \
  /** @internal */static inline ENAME##_itor_t ENAME##_list_next(ENAME##_itor_t i) \
    { return i->_next; } \
  /** @internal */static inline int ENAME##_list_size(ENAME##_list_t l) \
    { ENAME##_itor_t i=l->_head; int k=0; while(i!=NULL){k++;i=i->_next;} return k; }

#define PUK_LIST_FIND(VAR, LIST, COND_TEST) \
  for((VAR)=(LIST)->_head; (VAR); (VAR)=(VAR)->_next){if((COND_TEST))break;}

#define PUK_LIST_FOR(VAR, LIST) \
  for((VAR)=(LIST)->_head; (VAR); (VAR)=(VAR)->_next)

#define puk_list_foreach(ITOR, LIST) \
  for((ITOR)=(LIST)->_head; (ITOR); (ITOR)=(ITOR)->_next)

/** @remarks Puk_list how-to
 * PUK_LIST_DECLARE_TYPE(FOO) declares a FOO_t type and a FOO_list_t type
 * FOO_t is an element, FOO_list_t is a list of FOO_t elements
 * PUK_LIST_CREATE_TYPE(FOO, declrations) creates FOO_t and FOO-list_t types
 * and the related functions.
 * Example:
 * PUK_LIST_CREATE_TYPE(FOO, int a; float b;)
 * functions:
 *   FOO_t      FOO_new(void);
 *   void       FOO_delete(FOO_t);
 *   FOO_list_t FOO_list_new(void);
 *   void       FOO_list_delete(FOO_list_t);
 *   int        FOO_list_empty(FOO_list_t);
 *   void       FOO_list_erase(FOO_list_t, FOO_t);
 *   void       FOO_list_push_back(FOO_list_t, FOO_t);
 *   void       FOO_list_push_front(FOO_list_t, FOO_t);
 *   FOO_t      FOO_list_back(FOO_list_t);
 *   FOO_t      FOO_list_front(FOO_list_t);
 */

/** @todo
 * pop_back()
 * size()
 */


