/** @file
 * @brief Puk module management
 */

/** @defgroup PukMod API: Puk modules -- module management
 * @ingroup Puk
 */

  /** @addtogroup PukMod
   * @{
   */

  typedef int  (*puk_init_fun_t)(void);
  typedef int  (*puk_run_fun_t)(int, char**);
  typedef void (*puk_finalize_fun_t)(void);

  typedef struct padico_mod_s*padico_mod_t;

  typedef const char*padico_modID_t;
  PUK_VECT_TYPE(padico_modID, padico_modID_t);

  struct padico_unit_s
  {
    char* name;            /**< name of the unit as in the xml description file */
    padico_mod_t mod;      /**< the owner mod of this unit */
    void* driver_specific; /**< info-block for driver-dependant use */
  };
  typedef struct padico_unit_s* padico_unit_t;

  typedef void (*puk_notifier_t)(void*);
  /** a Puk job descriptor */
  struct puk_job_s
  {
    int            argc;       /**< argument count */
    char**         argv;       /**< arguments value */
    puk_notifier_t notify;     /**< notifier called at the end of the job */
    void*          notify_key; /**< key given to the notifier */
    padico_rc_t    rc;         /**< return code of the job */
    void*          driver_specific;
  };
  typedef struct puk_job_s*puk_job_t;

  /** A module driver */
  struct padico_loader_s
  {
    const char*name;                                  /**< name for the driver */
    padico_rc_t (*load)   (padico_unit_t);            /**< function able to load one unit */
    padico_rc_t (*start)  (padico_unit_t, puk_job_t); /**< function to start a new job */
    padico_rc_t (*unload) (padico_unit_t);            /**< function to unload one unit */
  };
  typedef const struct padico_loader_s*puk_driver_t;

  struct puk_lib_s
  {
    char* name;
    char* filename;
    void* dl_handle;
  };
  PUK_AUTO_TYPE(puk_lib);
  padico_rc_t   puk_lib_open   (puk_lib_t* lib, const char*lib_name);
  void          puk_lib_setself(puk_lib_t lib, padico_mod_t mod);
  int           puk_lib_init   (puk_lib_t lib);
  puk_run_fun_t puk_lib_run_resolve (puk_lib_t lib);
  padico_rc_t   puk_lib_unload (puk_lib_t lib);


  padico_rc_t  padico_puk_mod_open     (padico_mod_t*, const char*mod_name);
  padico_rc_t  padico_puk_mod_open_file(padico_mod_t*mod, const char*file_name);
  padico_rc_t  padico_puk_mod_resolve  (padico_mod_t*, const char*);
  padico_rc_t  padico_puk_mod_load     (padico_mod_t);
  padico_rc_t  padico_puk_mod_start    (padico_mod_t, puk_job_t job);
  padico_rc_t  padico_puk_mod_unload   (padico_mod_t);
  void         padico_puk_driver_add   (const struct padico_loader_s*);
  padico_modID_vect_t puk_mod_getmodIDs(void);
  /** @todo padico_puk_driver_remove() maybe? :-) */

  void            puk_job_notify(puk_job_t);

  padico_modID_t  puk_mod_getname(padico_mod_t mod);
  void            puk_mod_set_ancestor(padico_mod_t mod, padico_mod_t ancestor);
  padico_mod_t    puk_mod_getbyname(const char*mod_name);
  void            puk_mod_setattr(padico_mod_t, const char*label, const char*value);
  const char*     puk_mod_getattr(padico_mod_t, const char*label);
  const char*     puk_mod_local_getattr(padico_mod_t mod, const char*label);
  void            puk_mod_attr2argv(padico_mod_t mod, int *pargc, const char*const**pargv);
  padico_string_t puk_mod_attrs_dump(padico_mod_t mod);
  padico_string_t puk_mod_refs_dump(padico_mod_t mod);

  /** @} */

