/** @file PukABI symbols declaration
 * @warning to prevent conflicts from minor variations in function
 * prototypes, this file must include no header from standard libc.
 * @note this file is not supposed to be included directly
 */

#ifndef PADICO_BUILD_STRING
#include "padico_config.h"
#endif

#if(!defined(MKSYMFX) || !defined(MKSYMPX) || !defined(MKSYMNOSUPPORT))
#error "Puk-ABI-symbols.h included in bad context: required macros no defined."
#endif

/* *** PukABI toolbox ************************************** */

#define NOPREPROCESSING    /**< no pre-processing */
#define NOPOSTPROCESSING   /**< no post-processing */
#define ALWAYSREAL       1 /**< always use the real symbol */
#define ALWAYSVIRTUAL    0 /**< always use the virtual symbol */
/** use virtual if available, real else */
#define DEFAULTTOREAL(SYM) ((PUK_SYM_DESCNAME(SYM)).ptr_virtual == NULL)
/** chose virtual/real depending on the given file descriptor */
#define FOLLOWFD(FD) (puk_fds_isreal(FD))
#define FDFROMABI(FD) {(FD)=puk_fds_from_abi(FD);}
#define VIRTUALIZEFD(FD) \
     { if(((FD) > -1) && _will_use_real) { (FD) = puk_fds_alloc(1, (FD), __FUNCTION__); } }

/* *** errno */
#define PUK_ABI_MKERRNO(SYM) PUK_ABI_MKTYPE(SYM, int*, (void))
#if defined(PADICO_SYMBOL_ERRNO)
PUK_ABI_MKERRNO(PADICO_SYMBOL_ERRNO);
#else
PUK_ABI_MKERRNO(puk_abi_errno_sym);
#endif

/* *** fixed size integers */
#if (SIZEOF_INT == 8)
#  define __puk_abi_int64_t int
#elif (SIZEOF_LONG == 8)
#  define __puk_abi_int64_t long
#elif (SIZEOF_LONG_LONG == 8)
#  define __puk_abi_int64_t long long
#else
#  error "PukABI: cannot find 64 bits integers"
#endif
#if (SIZEOF_INT == 4)
#  define __puk_abi_int32_t int
#elif (SIZEOF_LONG == 4)
#  define __puk_abi_int32_t long
#elif (SIZEOF_LONG_LONG == 4)
#  define __puk_abi_int32_t long long
#else
#  error "PukABI: cannot find 32 bits integers"
#endif

// void*malloc(size_t);
MKSYMFX(malloc,  void*, (unsigned long size), (size), 
	NOPREPROCESSING, NOPOSTPROCESSING, DEFAULTTOREAL(malloc));
/*
  {puk_abi_malloc_count++;puk_abi_print("malloc()- size=%d (%d:%d)",size,puk_abi_malloc_count,puk_abi_malloc_count-puk_abi_free_count);},
  {puk_abi_print("malloc()- exit ptr=%X",result);}, 
  DEFAULTTOREAL(malloc));
*/

// void free(void*ptr);
MKSYMPX(free,    void,  (void*ptr), (ptr), 
	NOPREPROCESSING, NOPOSTPROCESSING, DEFAULTTOREAL(free));
/*
  {puk_abi_free_count++; puk_abi_print("free()- ptr=%X (%d:%d)",ptr, puk_abi_free_count,puk_abi_malloc_count-puk_abi_free_count);}, 
  {puk_abi_print("free()- exit");},
  DEFAULTTOREAL(free));
*/
// void*realloc(void*ptr, size_t size);
MKSYMFX(realloc, void*, (void *ptr, unsigned long size), (ptr, size),
	NOPREPROCESSING, NOPOSTPROCESSING, DEFAULTTOREAL(realloc));
/*
  {puk_abi_print("realloc()- size=%d ptr=%X",size, ptr);},
  {puk_abi_print("realloc()- exit ptr=%X",result);},
  DEFAULTTOREAL(realloc));
*/
// void*calloc(size_t nelem, size_t elsize);
MKSYMFX(calloc,  void*, (unsigned long nelem, unsigned long elsize), (nelem, elsize),
	NOPREPROCESSING, NOPOSTPROCESSING, DEFAULTTOREAL(calloc));
/*
  {puk_abi_print("calloc()- nelem=%d; elsize=%d",nelem,elsize);},
  {puk_abi_print("calloc()- exit ptr=%X",result);},
  DEFAULTTOREAL(calloc));
*/

// char *strdup(const char *s);
MKSYMFX(strdup,  char*, (const char*s), (s),
	NOPREPROCESSING, NOPOSTPROCESSING, DEFAULTTOREAL(strdup));

// int socket(int domain, int type, int protocol)
MKSYMFX(socket,  int, (int domain, int type, int protocol), (domain, type, protocol),
	NOPREPROCESSING,
	VIRTUALIZEFD(result),
	DEFAULTTOREAL(socket));

// int flock(int fd, int operation);
MKSYMFX(flock, int, (int fd, int op), (fd, op),
	{ puk_abi_print("flock()fd=%d", fd); FDFROMABI(fd)},
	NOPOSTPROCESSING,
	FOLLOWFD(fd));

// int open(const char *path, int oflag, /* mode_t mode */...);
MKSYMVX(open, int, (const char*path, int oflag, ...), path, oflag,
{
  va_list _ap;
  va_start(_ap, oflag);
  const int o = puk_abi_oracle__open(oflag);
  if(o)
    {
      unsigned mode = va_arg(_ap, unsigned);
      puk_abi_trace("PukABI: open()- path=%s; mode=%d", path, (int)mode);
      _result = (*_fun)(path, oflag, mode);
    }
  else
    {
      puk_abi_trace("PukABI: open()- path=%s", path);
      _result = (*_fun)(path, oflag);
    }
  puk_abi_trace("PukABI: open()- result=%d", _result);
  VIRTUALIZEFD(_result);
  va_end(_ap);
},
  DEFAULTTOREAL(open));
// int open64(const char *path, int oflag, /* mode_t mode */...);
MKSYMVX(open64, int, (const char*path, int oflag, ...), path, oflag,
{
  va_list _ap;
  va_start(_ap, oflag);
  const int o = puk_abi_oracle__open(oflag);
  if(o)
    {
      unsigned mode = va_arg(_ap, unsigned);
      puk_abi_trace("PukABI: open64()- fun=%p; path=%s; flags=%d; mode=%d", _fun, path, oflag, (int)mode);
      _result = (*_fun)(path, oflag, mode);
    }
  else
    {
      puk_abi_trace("PukABI: open64()- fun=%p; path=%s; flags=%d", _fun, path, oflag);
      _result = (*_fun)(path, oflag);
    }
  puk_abi_trace("PukABI: open64()- result=%d", _result);
  VIRTUALIZEFD(_result);
  va_end(_ap);
},
  DEFAULTTOREAL(open64));
// int close(int fd);
MKSYMFX(close,   int, (int fd), (fd),
        {int save_fd=fd;fd=puk_fds_from_abi(fd);, if(result==0)puk_fds_free(save_fd);}, FOLLOWFD(fd));
// int dup(int fd);
MKSYMFX(dup, int, (int fd), (fd),
	FDFROMABI(fd),
	VIRTUALIZEFD(result), 
	DEFAULTTOREAL(dup));
// int dup2(int fd1, int fd2);
MKSYMVX(dup2, int, (int fd1, int fd2,...), fd1, fd2,
{ /* copy fd1 to fd2 */
  if(!puk_fds_isreal(fd1))
    puk_abi_fatal("PukABI: dup2()- fd1=%d; fd2=%d; fd1 is not real. Not supported yet.\n", fd1, fd2);
  if(puk_abi_fds[fd2].kind != PUK_FDS_NONE)
    puk_abi_fatal("PukABI: dup2()- fd1=%d; fd2=%d; fd2 is already in use. Not supported yet.\n", fd1, fd2);
  /* fd1 is real and fd2 is empty */
  const int _real_fd1 = puk_fds_from_abi(fd1);
  const int _real_fd2 = (*(dup_func_t)(PUK_SYM_REAL(dup)))(_real_fd1);
  (void)_fun; /* fake the use of '_fun' to avoid a warning */
  if(_real_fd2 > -1)
  {
    puk_abi_trace("PukABI: dup2()- fd1=%d; real_fd1=%d; register real_fd2=%d as abi_fd2=%d",
                  _real_fd1, fd1, _real_fd2, fd2);
    puk_fd_register(fd2, _real_fd2, PUK_FDS_REAL);
  }
  else
    puk_abi_fatal("PukABI: dup2()- cannot dup() real fd=%d\n", _real_fd1);
},
  DEFAULTTOREAL(dup2));

// int bind(int s, const struct sockaddr *name, int namelen);
MKSYMFX(bind, int, (int s, const void*name, unsigned int namelen), (s, name, namelen),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));
// int listen(int s, int backlog);
MKSYMFX(listen, int, (int s, int backlog), (s, backlog),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));
// int connect(int s, const struct sockaddr*name, socklen_t namelen);
MKSYMFX(connect, int, (int s, const void*name, unsigned int namelen), (s, name, namelen),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));
// int accept(int s, struct sockaddr*addr, socklen_t*addrlen);
MKSYMFX(accept, int, (int s, void*addr, unsigned int*addrlen), (s, addr, addrlen),
  FDFROMABI(s),
  {if(result>=0){VIRTUALIZEFD(result);}}, FOLLOWFD(s));
// int shutdown(int s, int how);
MKSYMFX(shutdown, int, (int s, int how), (s, how),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));
// ssize_t read(int fildes, void *buf, size_t nbyte);
MKSYMFX(read, int, (int fd, void*buf, int nbytes), (fd, buf, nbytes),
  FDFROMABI(fd), 
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// ssize_t pread(int fildes, void *buf, size_t nbyte, off_t offset);
MKSYMFX(pread, int, (int fd, void*buf, int nbytes, int offset), (fd, buf, nbytes, offset),
  FDFROMABI(fd), 
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// ssize_t write(int fildes, const void *buf, size_t nbyte)
MKSYMFX(write, int, (int fd, const void*buf, int nbytes), (fd, buf, nbytes),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// ssize_t pwrite(int fildes, const void *buf, size_t nbyte, off_t offset)
MKSYMFX(pwrite, int, (int fd, const void*buf, int nbytes, int offset), (fd, buf, nbytes, offset),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// ssize_t send(int s, const void *msg, size_t len, int flags);
MKSYMFX(send, int, (int s, const void*msg, unsigned len, int flags), (s, msg, len, flags),
  FDFROMABI(s), 
  NOPOSTPROCESSING,
  FOLLOWFD(s));
// ssize_t recv(int s, void *buf, size_t len, int flags);
MKSYMFX(recv, int, (int s, void*buf, unsigned len, int flags), (s, buf, len, flags),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));
// ssize_t sendto(int s,  const  void  *msg,  size_t  len,  int  flags, const struct sockaddr *to, socklen_t tolen);
MKSYMFX(sendto, int, (int s, const void*msg, unsigned len, int flags, const void*to, unsigned int tolen),
  (s, msg, len, flags, to, tolen),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));
//  ssize_t recvfrom(int s, void *buf, size_t  len,  int  flags,  struct sockaddr *from, socklen_t *fromlen);
MKSYMFX(recvfrom, int, (int s, void*buf, unsigned len, int flags, void*to, unsigned int*tolen),
  (s, buf, len, flags, to, tolen),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));
// ssize_t readv(int fildes, const struct iovec *iov, int iovcnt);
MKSYMFX(readv, int, (int fd, void*iov, int iovcount), (fd, iov, iovcount),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// ssize_t writev(int fildes,  const  struct  iovec  *iov,  int iovcnt);
MKSYMFX(writev, int, (int fd, void*iov, int iovcount), (fd, iov, iovcount),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// ssize_t sendmsg(int s, const struct msghdr *msg, int flags);
MKSYMFX(sendmsg, int, (int fd, void*msg, int flags), (fd, msg, flags),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// ssize_t recvmsg(int s, struct msghdr *msg, int flags);
MKSYMFX(recvmsg, int, (int fd, void*msg, int flags), (fd, msg, flags),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
//void *mmap(void *addr, size_t len, int prot, int flags,  int  fildes, off_t off);
MKSYMFX(mmap, void*, (void*addr, int len, int prot, int flags, int fd, int off), (addr, len, prot, flags, fd, off),
  { fd=(fd==-1)?-1:puk_fds_from_abi(fd); puk_abi_trace("PukABI: mmap()- addr=%p; len=%d; fd=%d", addr, len, fd);},
  { puk_abi_trace("PukABI: mmap()- rc=%p", result); }/* NOPOSTPROCESSING */, 
  ((fd==-1)?(DEFAULTTOREAL(mmap)):(FOLLOWFD(fd))));
//int munmap(void *addr, size_t len);
MKSYMFX(munmap, int, (void*addr, int len), (addr, len),
  NOPREPROCESSING,
  NOPOSTPROCESSING,
  DEFAULTTOREAL(munmap));
// int fileno(FILE *stream);
MKSYMFX(fileno, int, (void*stream), (stream),
  NOPREPROCESSING, 
{
  if(result>2&&(puk_fds_lookup(result, 1)==-1)) { VIRTUALIZEFD(result); }
}, ALWAYSREAL);
// int fcntl(int fildes, int cmd, /* arg */ ...);
MKSYMVX(fcntl, int, (int fd, int cmd, ...), fd, cmd,
  {
    int oldfd = fd;
    va_list _ap;
    va_start(_ap, cmd);
    FDFROMABI(fd);
    puk_abi_trace("fcntl()- will_use_real=%d; fd=%d; oldfd=%d; cmd=%d; fcntl=0x%lx [%s]",
      _will_use_real, fd, oldfd, cmd, _fun, __FUNCTION__);
    const int o = puk_abi_oracle__fcntl(cmd);
    switch(o)
    {
      case 0:
        _result = (*_fun)(fd, cmd);
        break;
      case 1:
        {
          long a = va_arg(_ap, long);
          _result = (*_fun)(fd, cmd, a);
        }
        break;
      case 2:
        {
          void*ptr = va_arg(_ap, void*);
          _result = (*_fun)(fd, cmd, ptr);
        }
        break;
      default:
      _result = -1;
      break;
    }
    puk_abi_trace("fcntl()- rc=%d [%s]", _result, __FUNCTION__);
    va_end(_ap);
   },
  FOLLOWFD(fd));
// int getsockname(int  s,  struct  sockaddr  *name,  socklen_t *namelen);
MKSYMFX(getsockname, int, (int s, void*name, void*namelen), (s, name, namelen),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));
// int select(int nfds, fd_set*readfds, fd_set*writefds, fd_set*errorfds, struct timeval*timeout);
MKSYMFX(select, int, (int nfds, void*rfd, void*wfd, void*efd, void*to), (nfds, rfd, wfd, efd, to),
{
  puk_abi_trace("stub__select()- entering- nfds=%d", nfds);
  puk_fdset_from_abi(rfd, nfds);
  puk_fdset_from_abi(wfd, nfds);
  puk_fdset_from_abi(efd, nfds);
},
{
  puk_fdset_to_abi(rfd, nfds, _will_use_real);
  puk_fdset_to_abi(wfd, nfds, _will_use_real);
  puk_fdset_to_abi(efd, nfds, _will_use_real);
  puk_abi_trace("stub__select()- exiting- result=%d", result);
},
  puk_fdset_isreal(rfd, wfd, efd, nfds));
// int poll(struct pollfd fds[], nfds_t nfds, int timeout);
MKSYMFX(poll, int, (void*fds, unsigned int nfds, int timeout), (fds, nfds, timeout),
  {puk_pollfd_from_abi(fds, nfds);},
  {puk_pollfd_to_abi(fds, nfds);},
  ALWAYSVIRTUAL);

// int setsockopt(int s, int level,  int  optname,  const  void *optval, int optlen)
MKSYMFX(setsockopt, int, 
  (int s, int level, int optname, const void*optval, int optlen),
  (s, level, optname, optval, optlen),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));
// int getsockopt(int s, int level, int optname, void  *optval, int *optlen)
MKSYMFX(getsockopt, int,
  (int s, int level, int optname, void*optval, int*optlen),
  (s, level, optname, optval, optlen),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));
// off_t lseek(int fildes, off_t offset, int whence);
MKSYMFX(lseek, int, (int fd, int offset, int whence), (fd, offset, whence),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// off64_t lseek64(int fildes, off64_t offset, int whence);
MKSYMFX(lseek64, __puk_abi_int64_t, (int fd, __puk_abi_int64_t offset, int whence), (fd, offset, whence),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// int ftruncate(int fildes, off_t length);
MKSYMFX(ftruncate, int, (int fd, long length), (fd, length),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// int ftruncate64(int fildes, off64_t length);
MKSYMFX(ftruncate64, int, (int fd, __puk_abi_int64_t length), (fd, length),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// int creat(const char *path, mode_t mode);
MKSYMFX(creat, int, (const char*path, unsigned mode), (path, mode),
  NOPREPROCESSING, 
{
  if(result >= 0)
    { VIRTUALIZEFD(result); }
},
  DEFAULTTOREAL(creat));
// int fsync(int fd);
MKSYMFX(fsync, int, (int fd), (fd),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// int pipe(int fildes[2]);
MKSYMFX(pipe, int, (int fd[]), (fd),
  NOPREPROCESSING, 
{
  if(result==0) {
    VIRTUALIZEFD(fd[0]);
    VIRTUALIZEFD(fd[1]);
  }
}, DEFAULTTOREAL(pipe));
// int mkfifo(const char *pathname, mode_t mode);
MKSYMFX(mkfifo, int, (const char*pathname, unsigned mode), (pathname, mode),
  NOPREPROCESSING,
  { if(result >= 0) { VIRTUALIZEFD(result); } },
  DEFAULTTOREAL(mkfifo));
// int socketpair(int d, int type, int protocol, int sv[2]);
MKSYMFX(socketpair, int, (int fd, int type, int protocol, int sv[2]), (fd, type, protocol, sv),
  NOPREPROCESSING,
{
  if(result==0) {
    VIRTUALIZEFD(sv[0]);
    VIRTUALIZEFD(sv[1]);
  }
}, DEFAULTTOREAL(socketpair));
// int getpeername(int  s,  struct  sockaddr  *name,  socklen_t *namelen);
MKSYMFX(getpeername, int, (int s, void*name, int*namelen), (s, name, namelen),
  FDFROMABI(s),
  NOPOSTPROCESSING,
  FOLLOWFD(s));

//void exit(int status);
MKSYMPX(exit, void, (int status), (status),
  NOPREPROCESSING,
  NOPOSTPROCESSING,
  DEFAULTTOREAL(exit));

/*
PUK_ABI_MKTYPE(system, int, (char*s));
*/
// int system(const char *string);
MKSYMFX(system, int, (const char*cmd), (cmd),
  NOPREPROCESSING,
  NOPOSTPROCESSING,
  ALWAYSREAL);

#ifdef HAVE___FXSTAT
/** @note Typical GLIBC *statically* define 'fstat' as an alias to '__fxstat'. 
 * So we must override '__fxstat', not 'fstat'.
 */
// int __fxstat(int version, int fildes, struct stat *buf);
MKSYMFX(__fxstat, int, (int version, int fd, void*buf), (version, fd, buf),
  { FDFROMABI(fd); puk_abi_trace("PukABI: __fxstat() fd=%d", fd);  },
  { puk_abi_trace("PukABI: __fxstat()- rc=%d", result); } /* NOPOSTPROCESSING */,
  FOLLOWFD(fd));
// int __fxstat64(int version, int fildes, struct stat *buf);
MKSYMFX(__fxstat64, int, (int version, int fd, void*buf), (version, fd, buf),
  { FDFROMABI(fd); puk_abi_trace("PukABI: __fxstat64() fd=%d", fd);  },
  { puk_abi_trace("PukABI: __fxstat64()- rc=%d", result); } /* NOPOSTPROCESSING */,
  FOLLOWFD(fd));
#else /* HAVE___FXSTAT */
// int fstat(int fildes, struct stat *buf);
MKSYMFX(fstat, int, (int fd, void*buf), (fd, buf),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
#endif /* HAVE___FXSTAT */
/** @note not strictly required to wrap 'unlink' but needed for YAMPI to avoid EINTR.
 */
// int unlink(const char *pathname);
MKSYMFX(unlink, int, (const char*pathname), (pathname),
  NOPREPROCESSING,
  NOPOSTPROCESSING,
  ALWAYSREAL);

/** @todo We assume the form: 
 * ioctl(int fd, int req, unsigned long arg);
 * It is not a big issue, though, since it covers 99%
 * of ioctl typical usage.
 */
// int ioctl(int fildes, int request, /* arg */ ...);
MKSYMVX(ioctl, int, (int fd, int req, ...), fd, req,
{
  va_list _ap;
  va_start(_ap, req);
  FDFROMABI(fd);
  unsigned long v = va_arg(_ap, unsigned long);
  puk_abi_trace("PukABI: ioctl()- fd=%d", fd);
  _result = (*_fun)(fd, req, v);
  va_end(_ap);
},
  FOLLOWFD(fd));

#ifdef HAVE_INOTIFY_H
// int inotify_init(void)
MKSYMFX(inotify_init,  int, (void), (),
  NOPREPROCESSING,
  VIRTUALIZEFD(result),
  DEFAULTTOREAL(socket));
// int inotify_add_watch(int fd, const char *pathname, uint32_t mask);
MKSYMFX(inotify_add_watch, int, (int fd, const char*pathname, __puk_abi_int32_t mask), (fd, pathname, mask),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
// int inotify_rm_watch(int fd, uint32_t wd);
MKSYMFX(inotify_rm_watch, int, (int fd, __puk_abi_int32_t wd), (fd, wd),
  FDFROMABI(fd),
  NOPOSTPROCESSING,
  FOLLOWFD(fd));
#endif /* HAVE_INOTIFY_H */

//int execve(const char *filename, char *const argv[], char *const envp[]);
MKSYMFX(execve, int, (const char*name, char*const argv[], char*const arge[]), (name, argv, arge),
  NOPREPROCESSING,
  NOPOSTPROCESSING,
  DEFAULTTOREAL(execve));

//void usleep(unsigned long usec);
MKSYMPX(usleep, void, (unsigned long usec), (usec),
  NOPREPROCESSING,
  NOPOSTPROCESSING,
  DEFAULTTOREAL(usleep));

//unsigned int sleep(unsigned int seconds);
MKSYMFX(sleep, unsigned int, (unsigned int sec), (sec),
  NOPREPROCESSING,
  NOPOSTPROCESSING,
  DEFAULTTOREAL(sleep));

//int nanosleep(const struct timespec *req, struct timespec *rem)
MKSYMFX(nanosleep, int, (const void*req, void*rem), (req, rem),
  NOPREPROCESSING,
  NOPOSTPROCESSING,
  DEFAULTTOREAL(nanosleep));

MKSYMNOSUPPORT(sendfile);
MKSYMNOSUPPORT(sendfile64);
MKSYMNOSUPPORT(epoll_create);
MKSYMNOSUPPORT(epoll_ctl);
MKSYMNOSUPPORT(epoll_wait);


MKSYMNOSUPPORT(ulimit);
MKSYMNOSUPPORT(valloc);
MKSYMNOSUPPORT(pvalloc);


