/** @file
 * Provides the "boot" module driver (derived from "binary")
 * @ingroup Puk
 */

#include "Puk-internals.h"
#include "Module.h"

PADICO_MODULE_HOOK(Puk);


/* ********************************************************* */

static padico_rc_t puk_unit_boot_load  (padico_unit_t unit);
static padico_rc_t puk_unit_boot_unload(padico_unit_t unit);

static const struct padico_loader_s puk_boot_driver =
{
  .name   = "boot",
  .load   = &puk_unit_boot_load,
  .start  = NULL, /* no 'start'- the 'load' operation takes the control. */
  .unload = &puk_unit_boot_unload
};

/* ********************************************************* */

static padico_rc_t puk_unit_boot_load(padico_unit_t unit)
{
  padico_rc_t rc = NULL;
  puk_lib_t lib;
  if(unit->driver_specific)
    {
      padico_out(10, "%s (boot) already loaded -- skip\n", unit->name);
      return rc;
    }
  rc = puk_lib_open(&lib, unit->name);
  if(padico_rc_iserror(rc))
    {
      padico_warning("Puk: [boot] binary library load error:\n");
      padico_rc_show(rc);
      padico_fatal("Puk: [boot] cannot boot. Process halted.\n");
    }
  else
    {
      unit->driver_specific = (void*)lib;
      puk_lib_setself(lib, unit->mod);
      puk_lib_init(lib);
    }
  return rc;
}

static padico_rc_t puk_unit_boot_unload(padico_unit_t unit)
{
  const puk_lib_t lib = (puk_lib_t)unit->driver_specific;
  padico_rc_t rc = puk_lib_unload(lib);
  unit->driver_specific = NULL;
  return rc;
}

/* ********************************************************* */

void __PUK_SYM_INTERNAL padico_puk_boot_init(void)
{
  padico_puk_driver_add(&puk_boot_driver);
}

