/** @file
 * @brief Return codes and error management
 * @ingroup Puk
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "Puk-internals.h"

/******************************************************/
/* ** display functions*/

void padico_print(char*format, ...)
{
#ifdef PADICO_ENABLE_OUTPUT
  va_list ap;
#define BIG_ENOUGH 2048
  char buffer[BIG_ENOUGH];
  int i;
  va_start(ap, format);
  i = snprintf(buffer, BIG_ENOUGH, "%s: ", padico_hostname());
  i+= vsnprintf(&buffer[i], BIG_ENOUGH - i, format, ap);
  va_end(ap);
  fprintf(stderr, buffer);
  fflush(stderr);
#endif /* PADICO_ENABLE_OUTPUT */
}


/* *** the following functions are almost obsolete */

void padico_fatal(char*format, ...)
{
  va_list ap;
  va_start(ap, format);
  fprintf(stderr, "\n%s: [fatal] *** Padico fatal error ***\n", padico_hostname());
  fprintf(stderr, "%s: [fatal] ", padico_hostname());
  vfprintf(stderr, format, ap);
  va_end(ap);
  fprintf(stderr, "%s: [fatal] abort.\n\n", padico_hostname());
  abort();
}

void padico_warning(char*format, ...)
{
#ifdef PADICO_ENABLE_OUTPUT
  va_list ap;
  va_start(ap, format);
  fprintf(stderr, "%s: *** warning *** ", padico_hostname());
  vfprintf(stderr, format, ap);
  va_end(ap);
#endif /* PADICO_ENABLE_OUTPUT */
}

/******************************************************/

padico_rc_t padico_rc_create(int _rc, char*fmt, ...)
{
  padico_rc_t rc = padico_rc_new();
  int err;
  int restart;
  va_list ap;
  rc->rc = _rc;
  if(fmt && fmt[0])
    {
      rc->msg = padico_string_new();
      do {
	restart = 0;
	va_start(ap, fmt);
	err = vsnprintf(rc->msg->_text, rc->msg->_capacity, fmt, ap);
	if((err < 0) || (err >= rc->msg->_capacity))
	  {
	    padico_string_grow(rc->msg, err);
	    restart = 1;
	  }
	va_end(ap);
      }
      while(restart);
      rc->msg->_size = strlen(rc->msg->_text);
    }
  else
    rc->msg = NULL;
  return rc;
}

padico_rc_t padico_rc_cat(padico_rc_t rc1 , padico_rc_t rc2)
{
  if(rc2 == NULL)
    {
      return rc1;
    }
  if(rc1 == NULL)
    {
      rc1 = padico_rc_new();
    }
  if(rc1->rc || rc2->rc)
    {
      rc1->rc = -1;
    }
  if(rc1->msg == NULL)
    {
      rc1->msg = padico_string_new();
    }
  if(rc2->msg != NULL)
    {
      padico_string_cat(rc1->msg, rc2->msg->_text);
    }
 return rc1;
}

padico_rc_t padico_rc_add(padico_rc_t rc1 , padico_rc_t rc2)
{
  padico_rc_t rc = padico_rc_cat(rc1, rc2);
  if(rc2)
    padico_rc_delete(rc2);
  return rc;
}

void padico_rc_show(padico_rc_t rc)
{
  if(rc == NULL)
    {
      padico_print("ok/0\n");
    }
  else if(rc->rc)
    {
      padico_print("err/%s\n", padico_string_get(rc->msg));
    }
  else if(rc->msg)
    {
      padico_print("ok/%s\n", padico_string_get(rc->msg));
    }
  else
    {
      padico_print("err/incorrect code\n");
    }
}

char*padico_rc_gettext(padico_rc_t rc)
{
  if(rc == NULL)
    {
      return "";
    }
  else if(rc->rc)
    {
      return padico_string_get(rc->msg);
    }
  else if(rc->msg)
    {
      return padico_string_get(rc->msg);
    }
  else
    {
      padico_fatal("err/incorrect code\n");
    }
  return NULL;
}
