
# ## basic tools
AC_DEFUN([AC_PADICO_BASE_TOOLS],
	 [
	   dnl --- true
	   AC_PATH_PROG(BIN_TRUE, true)
	   AC_SUBST(BIN_TRUE)
	
	   dnl --- xterm
	   if test "x" = "x$XTERM" ; then
	     AC_PATH_PROGS(xterm_path, [ xterm x-terminal-emulator rxvt Eterm konsole ])
	     XTERM="$xterm_path"
	   else
	     XTERM="$XTERM"
	   fi
	   AC_ARG_VAR(XTERM, [ The X Window terminal emulator ])
	   AC_DEFINE_UNQUOTED(XTERM, "${XTERM}", [ X Window terminal emulator ])
	   if test "x$XTERM" = "x"; then
	     AC_MSG_ERROR([xterm not found.])
 	   fi
	
	   dnl --- dialog
	   AC_PATH_PROGS(DIALOG, [ ${DIALOG} whiptail dialog cdialog gdialog ])
	   AC_ARG_VAR(DIALOG, [ The program used to display dialogs (e.g. whiptail) ])

	   dnl --- gdb
	   AC_PATH_PROGS(GDB, [ ${GDB} gdb ])
	   AC_ARG_VAR(GDB, [ The gdb debugger ])
	   AC_DEFINE_UNQUOTED(GDB, "${GDB}", [ The gdb debugger ])
	   if test "x$GDB" = "x"; then
	     AC_MSG_ERROR([gdb not found.])
 	   fi
	 ])

# ## check for GNU make
AC_DEFUN([AC_PADICO_GNU_MAKE],
	 [ AC_CACHE_CHECK( for GNU make,_cv_gnu_make_command,
                _cv_gnu_make_command='' ;
dnl Search all the common names for GNU make
                for a in "$MAKE" make gmake gnumake gnu-make ; do
                        if test -z "$a" ; then continue ; fi ;
                        if  ( sh -c "$a --version" 2> /dev/null | grep GNU  2>&1 > /dev/null ) ;  then
                                _cv_gnu_make_command=$a ;
                                break;
                        fi
                done ;
        ) ;
dnl If there was a GNU version, then set @ifGNUmake@ to the empty string, '#' otherwise
        if test  "x$_cv_gnu_make_command" != "x"  ; then
                MAKE=$_cv_gnu_make_command
                ifGNUmake='' 
        else
                ifGNUmake='#'
                AC_MSG_ERROR([Not found])
        fi
        AC_ARG_VAR(MAKE, [ path for GNU make ])
        AC_SUBST(ifGNUmake)
	 ] )

# ##  C/C++ dependency checker
AC_DEFUN([AC_PADICO_DEP],
	 [ AC_REQUIRE([AC_PADICO_COMPILER])
	   AC_MSG_CHECKING([how to check C dependency])
	   CC_DEP="${CC} ${CDEPFLAGS}"
	   AC_MSG_RESULT([${CC_DEP}])
	   AC_SUBST(CC_DEP)
	   AC_MSG_CHECKING([how to check C++ dependency])
	   CXX_DEP="${CXX} ${CXXDEPFLAGS}"
	   AC_MSG_RESULT([${CXX_DEP}])
	   AC_SUBST(CXX_DEP)
	 ])

# ## static linker
AC_DEFUN([AC_PADICO_LINKER],
	 [ AC_REQUIRE([AC_PADICO_COMPILER])
	   AC_MSG_CHECKING([for static linker])
	   LD="${CC}"
	   AC_MSG_RESULT([${LD} ${LDFLAGS}])
	   AC_SUBST(LD)
	   AC_SUBST(LDFLAGS)
	 ])

# ## dynamic linker
AC_DEFUN([AC_PADICO_DYNLD],
	 [ AC_REQUIRE([AC_PADICO_COMPILER])
	   AC_MSG_CHECKING([for dynamic library linker])
	   DYNLIB_LD="${CC} ${CPICFLAGS} ${DYNLIB_FLAGS}"
	   AC_MSG_RESULT([${DYNLIB_LD}])
	   AC_SUBST(DYNLIB_LD)

	   AC_MSG_CHECKING([for dynamic library prefix])
	   AC_MSG_RESULT([${DYNLIB_PREFIX}])
	   AC_SUBST(DYNLIB_PREFIX)
	   AC_DEFINE_UNQUOTED(DYNLIB_PREFIX, "${DYNLIB_PREFIX}",
	     [prefix prepended to dynamic library names (usually "lib")])

	   AC_MSG_CHECKING([for dynamic library extension])
	   AC_MSG_RESULT([${DYNLIB_EXT}])
	   AC_SUBST(DYNLIB_EXT)
	   AC_DEFINE_UNQUOTED(DYNLIB_EXT, "${DYNLIB_EXT}",
	     [suffix for dynamic library names (usually ".so" on Unix systems)])
	 ])

# ## dynamic objects
AC_DEFUN([AC_PADICO_DSO],
	 [ AC_REQUIRE([AC_PADICO_COMPILER])
	   AC_MSG_CHECKING([for dynamic object (DSO) linker])
	   DSO_LD="${CC} ${CPICFLAGS} ${DSO_FLAGS}"
	   AC_MSG_RESULT([${DSO_LD}])
	   AC_SUBST(DSO_LD)

	   AC_MSG_CHECKING([for DSO extension])
	   AC_MSG_RESULT([${DSO_EXT}])
	   AC_SUBST(DSO_EXT)
	   AC_DEFINE_UNQUOTED(DSO_EXT, "${DSO_EXT}",
	     [Extension for bundles])

	   AC_MSG_CHECKING([for symbol prefix])
	   AC_MSG_RESULT([${SYMBOL_PREFIX}])
	   AC_SUBST(SYMBOL_PREFIX)
	   AC_DEFINE_UNQUOTED(SYMBOL_PREFIX, "${SYMBOL_PREFIX}",
	     [prefix prepended to symbols by the ABI (usually "_")])
	 ])

AC_DEFUN([AC_PADICO_EXPAT],
	 [ AC_MSG_CHECKING([for Expat XML parser])
	   AC_CHECK_HEADERS(expat.h, 
	     [ HAVE_EXPAT_H="yes" ],
	     [ HAVE_EXPAT_H="no"  ])
	   if test $HAVE_EXPAT_H = "no"; then
	     if test "x$EXPAT_ROOT" = "x"; then
	       AC_MSG_ERROR([ Expat XML parse not found ])
	     fi
	     AC_MSG_RESULT($EXPAT_ROOT)
	     CFLAGS="$CFLAGS -I$EXPAT_ROOT/include"
	     LDFLAGS="$LDFLAGS -L$EXPAT_ROOT/lib"
	     AC_SUBST(CFLAGS)
	     AC_SUBST(LDFLAGS)
	   else
	     AC_MSG_RESULT($HAVE_EXPAT_H)
	   fi
	 ])


	     
	
