#! /bin/echo "This script is not standalone. It is supposed to be included in another bash script."

if [ -r ${PADICO_ROOT}/etc/padico-install.conf ]; then
  . ${PADICO_ROOT}/etc/padico-install.conf
fi

# ##########################################################
# find dialog system

for d in ${DIALOG} whiptail dialog cdialog gdialog; do
  p=$( /usr/bin/which ${d} 2> /dev/null )
  err=$?
  if [ ${err} -eq 0 -a "x${p}" != "x" -a -x "${p}" ]; then
    DIALOG="${p}"
    break
  fi
done

if [ "x${DIALOG}" = "x" ]; then
  padico_dialog=padico_dialog_text
else
  if [ "x${padico_dialog}" = "x" ]; then
    padico_dialog=padico_dialog_whiptail
  fi
fi

# ##########################################################
# usage of padico_dialog_* functions:
#   --msgbox    <text>
#   --inputbox  <text> <init>
#   --menu      <text> <tag> <item> [<tag> <item>]...
#   --checklist <text> <tag> <item> <val> [<tag> <item> <val>]...

padico_dialog_text() {
  title="$1"
  action="$2"
  text="$3"
  shift 3
  echo              1>&3
  echo "* ${boldtext}${title}${normaltext}" 1>&3
  echo "* -----"            1>&3
  case "${action}" in
    --msgbox)
      echo "${text}" 1>&3
    ;;
    --inputbox)
      default_answer="$1"
      echo "${text} [ ${boldtext}${default_answer}${normaltext} ] " 1>&3
      read answer
      if [ "x${answer}" = "x" ]; then
        answer="${default_answer}"
      fi
      echo "${answer}"
    ;;
    --menu)
      echo "${text}" 1>&3
      echo           1>&3
      max=0
      j=0
      for w in "$@"; do
        if [ $((j%2)) == 0 -a ${#w} -gt ${max} ]; then max=${#w}; fi
	j=$((${j}+1));
      done
      i=1
      while [ "x$1" != "x" ]; do
	tag="$1"
	item="$2"
	tags[i]="$tag"
	space="                            "
	padding=${space:0:${max}-${#tag}}
	echo -e " (${i})\t${tag}${padding} - ${item}" 1>&3
	i=$((${i}+1))
	shift 2
      done
      echo 1>&3
      echo "Your choice: [ ${boldtext}1${normaltext} ] " 1>&3
      read answer
      if [ "x${answer}" = "x" ]; then
	answer=1
      fi
      if [ "$answer" -ge 1 -a "$answer" -lt $i ]; then
	echo "${tags[$answer]}"
      else
	return 1
      fi
    ;;
    --checklist)
      echo "${text}" 1>&3
      echo           1>&3
      options=
      while [ "x$1" != "x" ]; do
	tag="$1"
	item="$2"
	val="$3"
	if [ ${val} = 0 ]; then default_answer="n"; else default_answer="y"; fi
	echo " - ${tag}: ${item}" 1>&3
	echo "Enable '${tag}'? [ ${boldtext}${default_answer}${normaltext} ]" 1>&3
	read answer
	if [ "x${answer}" = "x" ]; then
	  answer="${default_answer}"
	fi
	case ${answer} in
	  y|Y) options="${options} ${tag}" ; shift 3 ;;
	  n|N) shift 3 ;;
	  *) echo "# Invalid choice: ${answer}" 1>&3
	esac
      done
      echo "${options}"
    ;;
  esac
  echo
}

padico_dialog_whiptail() {
  if [ "x${padico_backtitle}" = "x" ]; then
      padico_backtitle="$0"
  fi
  title="$1"
  action="$2"
  text="$3"
  shift 3
  case ${action} in
    --msgbox)
      ${DIALOG} --title "$title" --backtitle "$padico_backtitle" \
        ${action} "${text}" 0 0 2>&1 1>&3 3>&-
    ;;
    --inputbox)
      ${DIALOG} --title "$title" --backtitle "$padico_backtitle" \
        ${action} "${text}" 0 0 -- "$*" 2>&1 1>&3 3>&-
    ;;
    --menu|--checklist)
      set -o pipefail
      ${DIALOG} --title "$title" --backtitle "$padico_backtitle" \
        ${action} "${text}" 0 0 0 "$@" 2>&1 1>&3 3>&- | tr '"' ' '
    ;;
    *)
      exit 1
    ;;
  esac
  err=$?
  return $err
}

# open a copy of stdout on fd 3.
exec 3>&1

