
/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/*  THESE: INTERFACE START */
/****************************************************************
 * Certains commentaires sont tiquets par U/S/C/I signifiant
 * User       : utile aux threads applicatifs pour exploiter ces mcanismes
 * Start      : utile pour l'initialisation
 * Call-backs : utile pour le concepteur de call-backs
 * Internal   : ne devrait pas tre utilis hors de Marcel
 *
 * Ce ne sont que des indications, pas des contraintes
 ****************************************************************/

#section macros
/*[SC]****************************************************************
 * Rcupre l'adresse d'une structure englobante 
 * Il faut donner :
 * - l'adresse du champ interne
 * - le type de la structure englobante
 * - le nom du champ interne dans la structure englobante
 */
#define struct_up(ptr, type, member) \
	tbx_container_of(ptr, type, member)

#section types

/*[U]****************************************************************
 * Les types abstaits pour le serveur et les vnements
 */
/* Le serveur dfinit les call-backs et les paramtres de scrutation 
 * utiliser. Il va ventuellement regrouper les ressources
 * enregistres (en cas de scrutation active par exemple)
 */
typedef struct marcel_ev_server *marcel_ev_server_t;

/* Une requte dfinit une entit qui pourra recevoir un vnement que
 * l'on pourra attendre.  Diverses requtes d'un serveur pourront tre
 * groupes (afin de dterminer plus rapidement l'ensemble des tats
 * de chacune des requtes  chaque scrutation)
 */
typedef struct marcel_ev_req *marcel_ev_req_t;

/* Un thread peut attendre l'arrive effective d'un vnement
 * correspondant  une requte. Le thread est dschedul tant que la
 * requte n'est pas prte (que l'vnement n'est pas reu).
 */
typedef struct marcel_ev_wait *marcel_ev_wait_t;

#section macros
/*[S]****************************************************************
 * Initialisation d'un serveur
 */

/* Initialisation statique
 * var : la variable constante marcel_ev_server_t
 * name: une chane de caractres pour identifier ce serveur dans les
 *        messages de debug
 */
#define MARCEL_EV_SERVER_DEFINE(var, name) \
  struct marcel_ev_server ma_s_#var = MARCEL_EV_SERVER_INIT(ma_s_#var, name); \
  const MARCEL_EV_SERVER_DECLARE(var) = &ma_s_#var

#section functions
/* Idem, mais dynamique */
/* #ifndef __cplusplus */
void marcel_ev_server_init(marcel_ev_server_t server, char* name);
/* #endif */

/* Enregistrement des call-backs utilisables */
__tbx_inline__ static int marcel_ev_server_add_callback(marcel_ev_server_t server, 
						marcel_ev_op_t op,
						marcel_ev_callback_t *func);
/* Rglage des paramtres de scrutation */
int marcel_ev_server_set_poll_settings(marcel_ev_server_t server, 
				       unsigned poll_points,
				       unsigned poll_frequency);

/* Dmarrage du serveur
 * Il devient possible d'attendre des vnements
 */
int marcel_ev_server_start(marcel_ev_server_t server);

/* Arrt du serveur
 * Il est ncessaire de le rinitialiser pour le redmarrer
 */
int marcel_ev_server_stop(marcel_ev_server_t server);

/*  THESE: PROTO_EV_WAIT START */
/*[U]****************************************************************
 * Fonctions  l'usage des threads applicatifs
 */

#section types
/* Attribut pouvant tre attach aux vnements */
enum {
	/* Dsactive la requte lorsque survient l'occurrence
	 * suivante de l'vnement */
	MARCEL_EV_ATTR_ONE_SHOT=1,
	/* Ne rveille pas les threads en attente d'vnements du
	 * serveur (ie marcel_ev_server_wait())*/
	MARCEL_EV_ATTR_NO_WAKE_SERVER=2,
};

#section functions
int marcel_per_lwp_polling_register(int *data,
                                    int value_to_match,
                                    void (*func) (void *),
                                    void *arg);

void marcel_per_lwp_polling_proceed(void);

/* Un raccourci pratique des fonctions suivantes, utile si l'on ne
 * soumet la requte qu'une seule fois. Les oprations suivantes sont
 * effectues : initialisation, soumission et attente d'un
 * vnement avec ONE_SHOT positionn */
int marcel_ev_wait(marcel_ev_server_t server, marcel_ev_req_t req,
		   marcel_ev_wait_t wait, marcel_time_t timeout);

/* Initialisation d'un vnement
 * ( appeler en premier si l'on utilise autre chose que marcel_ev_wait) */
int marcel_ev_req_init(marcel_ev_req_t req);

/* Ajout d'un attribut spcifique  une requte */
int marcel_ev_req_attr_set(marcel_ev_req_t req, int attr);

/* Soumission d'une requte (le serveur PEUT commencer  scruter si cela
 * lui convient) */
int marcel_ev_req_submit(marcel_ev_server_t server, marcel_ev_req_t req);

/* Abandon d'une requte et retour des threads en attente sur cette
 * requte (avec le code de retour fourni) */
int marcel_ev_req_cancel(marcel_ev_req_t req, int ret_code);

/* Attente bloquante d'un vnement sur une requte dj enregistre */
int marcel_ev_req_wait(marcel_ev_req_t req, marcel_ev_wait_t wait, 
		       marcel_time_t timeout);

/* Attente bloquante d'un vnement sur une quelconque requte du serveur */
int marcel_ev_server_wait(marcel_ev_server_t server, marcel_time_t timeout);

/* Renvoie une requte survenue n'ayant pas l'attribut NO_WAKE_SERVER
 * (utile au retour de server_wait())
 *  l'abandon d'une requte (wait, req_cancel ou ONE_SHOT) la requte
 * est galement retire de cette file (donc n'est plus consultable) */
marcel_ev_req_t marcel_ev_get_success_req(marcel_ev_server_t server);

/*  THESE: PROTO_EV_WAIT STOP */
/* Exclusion mutuelle pour un serveur d'vnements
 *
 * - les call-backs sont TOUJOURS appels  l'intrieur de cette
 *   exclusion mutuelle.
 * - les call-backs BLOCK_ONE|ALL doivent relcher puis reprendre ce
 *   lock avant et aprs l'appel systme bloquant avec les deux fonctions
 *   prvues pour (marcel_ev_callback_*).
 * - les fonctions prcdentes peuvent tre appeles avec ou sans ce
 *   lock
 * - le lock est relch automatiquement par les fonctions d'attente
 *   (marcel_ev_*wait*())

 * - les call-backs et les rveils des threads en attente sur les
 *   vnements signals par le call-back sont atomiques (vis--vis de 
 *   ce lock)
 * - si le lock est pris avant les fonctions d'attente (ev_wait_*),
 *   la mise en attente est atomique vis--vis des call-backs
 *   ie: un call-back signalant l'vnement attendu rveillera cette
 *   attente
 * - si un vnement  la proprit ONE_SHOT, le dsenregistrement est
 *   atomic vis--vis du call-back qui a gnr l'vnement.
 */
int marcel_ev_lock(marcel_ev_server_t server);
int marcel_ev_unlock(marcel_ev_server_t server);
/* Pour BLOCK_ONE et BLOCK_ALL avant et aprs l'appel systme bloquant */
int marcel_ev_callback_will_block(marcel_ev_server_t server);
int marcel_ev_callback_has_blocked(marcel_ev_server_t server);

/* Appel forc de la fonction de scrutation */
void marcel_ev_poll_force(marcel_ev_server_t server);
/* Idem, mais on est sr que l'appel a t fait quand on retourne */
void marcel_ev_poll_force_sync(marcel_ev_server_t server);

#section macros
/*[S]****************************************************************
 * Les constantes pour le polling
 * Elles peuvent tre ORed
 */
#define MARCEL_EV_POLL_AT_TIMER_SIG  1
#define MARCEL_EV_POLL_AT_YIELD      2
#define MARCEL_EV_POLL_AT_LIB_ENTRY  4
#define MARCEL_EV_POLL_AT_IDLE       8

#section types
/*[SC]****************************************************************
 * Les diffrents oprations call-backs possibles
 */
typedef enum {
	/* server, XX, req to poll, nb_req_grouped, flags */
	MARCEL_EV_FUNCTYPE_POLL_POLLONE,
	/* server, XX, NA, nb_req_grouped, NA */
	MARCEL_EV_FUNCTYPE_POLL_GROUP,
	/* server, XX, NA, nb_req_grouped, NA */
	MARCEL_EV_FUNCTYPE_POLL_POLLANY,
	/* Les suivants ne sont pas encore utiliss... */
	/* server, XX, req to block, NA, NA */
	MARCEL_EV_FUNCTYPE_BLOCK_WAITONE,
	MARCEL_EV_FUNCTYPE_BLOCK_WAITONE_TIMEOUT,
	MARCEL_EV_FUNCTYPE_BLOCK_GROUP,
	MARCEL_EV_FUNCTYPE_BLOCK_WAITANY,
	MARCEL_EV_FUNCTYPE_BLOCK_WAITANY_TIMEOUT,
	MARCEL_EV_FUNCTYPE_UNBLOCK_WAITONE,
	MARCEL_EV_FUNCTYPE_UNBLOCK_WAITANY,
	/* PRIVATE */
	MA_EV_FUNCTYPE_SIZE
} marcel_ev_op_t;

/*[C]****************************************************************
 * Le prototype des call-back
 * server : le serveur
 * op : l'opration (call-back) demande
 * req : pour *(POLL|WAIT)ONE* : la requte  tester en particulier
 * nb_req : pour POLL_* : le nombre de requtes groupes
 * option : flags dpendant de l'opration
 *  - pour POLL_POLLONE : 
 *     + EV_IS_GROUPED : si la requte est dj groupe
 *     + EV_ITER : si POLL_POLLONE est appele sur toutes les requtes
 *                 en attente (ie POLL_POLLANY n'est pas disponible)
 *
 * La valeur de retour est pour l'instant ignore.
 */
typedef int (marcel_ev_callback_t)(marcel_ev_server_t server, 
				   marcel_ev_op_t op,
				   marcel_ev_req_t req, 
				   int nb_ev, int option);

typedef marcel_ev_callback_t *marcel_ev_pcallback_t;

/*[C]****************************************************************
 * Les flags des call-backs (voir ci-dessus)
 */
enum {
	/* Pour POLL_POLLONE */
	MARCEL_EV_OPT_REQ_IS_GROUPED=1,
	MARCEL_EV_OPT_REQ_ITER=2,
};

#section macros
/*[C]****************************************************************
 * Itrateurs pour les call-backs
 */

/****************************************************************
 * Itrateur pour les requtes enregistres d'un serveur
 */

/* Itrateur avec le type de base
   marcel_ev_req_t req : itrateur
   marcel_ev_server_t server : serveur
*/
#define FOREACH_REQ_REGISTERED_BASE(req, server) \
  list_for_each_entry((req), &(server)->list_req_registered, chain_req_registered)

/* Idem mais protg (usage interne) */
#define FOREACH_REQ_REGISTERED_BASE_SAFE(req, tmp, server) \
  list_for_each_entry_safe((req), (tmp), &(server)->list_req_registered, chain_req_registered)

/* Itrateur avec un type utilisateur
   [User Type] req : pointeur sur structure contenant un struct marcel_req
                     (itrateur)
   marcel_ev_server_t server : serveur
   member : nom de struct marcel_ev dans la structure pointe par req
*/
#define FOREACH_REQ_REGISTERED(req, server, member) \
  list_for_each_entry((req), &(server)->list_req_registered, member.chain_req_registered)

/****************************************************************
 * Itrateur pour les requtes groupes de la scrutation (polling)
 */

/* Itrateur avec le type de base
   marcel_ev_req_t req : itrateur
   marcel_ev_server_t server : serveur
*/
#define FOREACH_REQ_POLL_BASE(req, server) \
  list_for_each_entry((req), &(server)->list_req_poll_grouped, chain_req_grouped)

/* Idem mais protg (usage interne) */
#define FOREACH_REQ_POLL_BASE_SAFE(req, tmp, server) \
  list_for_each_entry_safe((req), (tmp), &(server)->list_req_poll_grouped, chain_req_grouped)

/* Itrateur avec un type utilisateur
   [User Type] req : pointeur sur structure contenant un struct marcel_req
                     (itrateur)
   marcel_ev_server_t server : serveur
   member : nom de struct marcel_ev dans la structure pointe par req
*/
#define FOREACH_REQ_POLL(req, server, member) \
  list_for_each_entry((req), &(server)->list_req_poll_grouped, member.chain_req_grouped)


/****************************************************************
 * Itrateur pour les requtes groupes de l'attente bloquante
 */

/* Itrateur avec le type de base
   marcel_ev_req_t req : itrateur
   marcel_ev_server_t server : serveur
*/
#define FOREACH_REQ_BLOCK_BASE(req, server) \
  list_for_each_entry((req), &(server)->list_req_block, chain_req)

/* Idem mais protg (usage interne) */
#define FOREACH_REQ_BLOCK_BASE_SAFE(req, tmp, server) \
  list_for_each_entry_safe((req), (tmp), &(server)->list_req_block, chain_req)

/* Itrateur avec un type utilisateur
   [User Type] req : pointeur sur structure contenant un struct marcel_req
                     (itrateur)
   marcel_ev_server_t server : serveur
   member : nom de struct marcel_ev dans la structure pointe par req
*/
#define FOREACH_REQ_BLOCK(req, server, member) \
  list_for_each_entry((req), &(server)->list_req_block, member.chain_req)


/****************************************************************
 * Itrateur pour les attentes d'vnements rattachs  une requte
 */

/* Itrateur avec le type de base
   marcel_ev_wait_t wait : itrateur
   marcel_ev_req_t req : requte
*/
#define FOREACH_WAIT_BASE(wait, req) \
  list_for_each_entry((wait), &(req)->list_wait, chain_wait)

/* Idem mais protg (usage interne) */
#define FOREACH_WAIT_BASE_SAFE(wait, tmp, req) \
  list_for_each_entry_safe((wait), (tmp), &(req)->list_wait, chain_wait)

/* Itrateur avec un type utilisateur
   [User Type] req : pointeur sur structure contenant un struct marcel_req
                     (itrateur)
   marcel_ev_server_t server : serveur
   member : nom de struct marcel_ev dans la structure pointe par req
*/
#define FOREACH_WAIT(wait, req, member) \
  list_for_each_entry((wait), &(req)->list_wait, member.chain_wait)



/* Macro utilisable dans les call-backs pour indiquer qu'une requte
   a reu un vnement.

   * la requte sera fournie  marcel_ev_get_success_req() (tant
   qu'elle n'est pas dsenregistre)
   * les threads encore en attente d'un vnement li  cette requte seront
   rveills et renverront le code 0

   marcel_ev_req_t req : requte
*/
#define MARCEL_EV_REQ_SUCCESS(req) \
  do { \
        list_move(&(req)->chain_req_ready, &(req)->server->list_req_ready); \
  } while(0)
 


/* Macro utilisable dans les call-backs pour rveiller un thread en attente
   en lui fournissant le code de retour
   marcel_ev_wait_t wait : l'vnement  reveiller
*/
#define MARCEL_EV_WAIT_SUCCESS(wait, code) \
  do { \
        list_del(&(wait)->chain_wait); \
        (wait)->ret_code=(code); \
  } while(0)
 
/*  THESE: INTERFACE STOP */


/*  =============== PRIVATE =============== */
/****************************************************************
 * Ce qui suit n'a pas vocation  tre utilis hors de Marcel
 */

#section structures
#depend "[marcel_structures]"

#section inline
/*  #ifndef __cplusplus */
/*  __tbx_inline__ static void marcel_ev_server_init(marcel_ev_server_t server, char* name) */
/*  { */
/*  	*server=(struct marcel_ev_server)MARCEL_EV_SERVER_INIT(*server, name); */
/*  } */
/*  #endif */

__tbx_inline__ static int marcel_ev_server_add_callback(marcel_ev_server_t server, 
						marcel_ev_op_t op,
						marcel_ev_pcallback_t func)
{
#ifdef MA__DEBUG
	/* Cette fonction doit tre appele entre l'initialisation et
	 * le dmarrage de ce serveur d'vnements */
	MA_BUG_ON(server->state != MA_EV_SERVER_STATE_INIT);
	/* On vrifie que l'index est correct */
	MA_BUG_ON(op>=MA_EV_FUNCTYPE_SIZE || op<0);
	/* On vrifie que la fonction n'est pas dj l */
	MA_BUG_ON(server->funcs[op]!=NULL);
#endif
	server->funcs[op]=func;
	return 0;
}

#section functions
/* Utilis dans les initialiseurs */
void marcel_poll_from_tasklet(unsigned long id);
void marcel_poll_timer(unsigned long id);

#section marcel_structures
#depend "pm2_list.h"
#depend "marcel_sem.h[]"
#depend "marcel_descr.h[types]"
#depend "linux_interrupt.h[]"
#depend "linux_timer.h[]"

#section marcel_types
/****************************************************************
 * tat d'un serveur d'vnements
 */
enum {
	/* Non initialis */
	MA_EV_SERVER_STATE_NONE,
	/* En cours d'initialisation */
	MA_EV_SERVER_STATE_INIT=1,
	/* En fonctionnement */
	MA_EV_SERVER_STATE_LAUNCHED=2,
	/* En arrt */
	MA_EV_SERVER_STATE_HALTED=3,
};

/****************************************************************
 * Structure serveur d'vnements
 */
#section marcel_structures
struct marcel_ev_server {
	/* Spinlock rgissant l'accs  cette structure */
	ma_spinlock_t lock;
	/* Thread propritaire du lock (pour le locking applicatif) */
	marcel_task_t *lock_owner;
	/* Liste des requtes soumises */
	struct list_head list_req_registered;
	/* Liste des requtes signales prtes par les call-backs */
	struct list_head list_req_ready;
	/* Liste des requtes pour ceux en attente dans la liste prcdente */
	struct list_head list_req_success;
	/* Liste des attentes en cours sur les vnements */
	struct list_head list_id_waiters;
	/* Spinlock rgissant l'accs  la liste prcdente */
	ma_spinlock_t req_success_lock;
	/* Polling events registered but not yet polled */
	int registered_req_not_yet_polled;

	/* Liste des requtes groupes pour le polling (ou en cours de groupage) */
	struct list_head list_req_poll_grouped;
	/* Nombre de requtes dans la liste prcdente */
	int req_poll_grouped_nb;

	/* Call-backs */
	marcel_ev_pcallback_t funcs[MA_EV_FUNCTYPE_SIZE];
	/* Points et frquence de scrutation */
	unsigned poll_points;
	unsigned frequency;
	/* Chaine des serveurs en cours de polling 
	   (state == 2 et tches en attente de polling) */
	struct list_head chain_poll;
	/* Tasklet et timer utilise pour la scrutation */
	struct ma_tasklet_struct poll_tasklet;
	struct ma_timer_list poll_timer;
	/* tat du serveur */
	int state;
	/* Nom (utilis pour le dbug) */
	char* name;
};

#section macros
#define MARCEL_EV_SERVER_DECLARE(var) \
  const marcel_ev_server_t var
#define MARCEL_EV_SERVER_INIT(var, sname) \
  { \
    .lock=MA_SPIN_LOCK_UNLOCKED, \
    .lock_owner=NULL, \
    .list_req_registered=LIST_HEAD_INIT((var).list_req_registered), \
    .list_req_ready=LIST_HEAD_INIT((var).list_req_ready), \
    .list_req_success=LIST_HEAD_INIT((var).list_req_success), \
    .list_id_waiters=LIST_HEAD_INIT((var).list_id_waiters), \
    .req_success_lock=MA_SPIN_LOCK_UNLOCKED, \
    .registered_req_not_yet_polled=0, \
    .list_req_poll_grouped=LIST_HEAD_INIT((var).list_req_poll_grouped), \
    .req_poll_grouped_nb=0, \
    .funcs={NULL, }, \
    .poll_points=0, \
    .frequency=0, \
    .chain_poll=LIST_HEAD_INIT((var).chain_poll), \
    .poll_tasklet= MA_TASKLET_INIT((var).poll_tasklet, \
                     &marcel_poll_from_tasklet, \
                     (unsigned long)(marcel_ev_server_t)&(var) ), \
    .poll_timer= MA_TIMER_INITIALIZER(marcel_poll_timer, 0, \
                   (unsigned long)(marcel_ev_server_t)&(var)), \
    .state=MA_EV_SERVER_STATE_INIT, \
    .name=sname, \
  }


#section marcel_types
enum {
	MARCEL_EV_STATE_OCCURED=1,
	MARCEL_EV_STATE_GROUPED=2,
	MARCEL_EV_STATE_ONE_SHOT=4,
	MARCEL_EV_STATE_NO_WAKE_SERVER=8,
	MARCEL_EV_STATE_REGISTERED=16,
};

#section marcel_structures
struct marcel_ev_req {
	/* Chaine des requtes soumises */
	struct list_head chain_req_registered;
	/* Chaine des requtes groupes en attente */
	struct list_head chain_req_grouped;
	/* Chaine des requtes signales OK par un call-back */
	struct list_head chain_req_ready;
	/* Chaine des requtes  signaler au serveur */
	struct list_head chain_req_success;
	/* tat */
	int state;
	/* Serveur attach */
	marcel_ev_server_t server;
	/* Liste des attentes en cours sur cette requte */
	struct list_head list_wait;
};

struct marcel_ev_wait {
	/* Chaine des vnements group en attente */
	struct list_head chain_wait;

	marcel_sem_t sem;
	/* 0: event
	   -ECANCELED: marcel_unregister called
	*/
	int ret;
	marcel_task_t *task;
};


#section marcel_variables
#include "tbx_compiler.h"
/* Liste des serveurs en cours de polling 
 * (state == 2 et tches en attente de polling) 
 */
extern TBX_EXTERN struct list_head ma_ev_list_poll;

#section marcel_functions
static __tbx_inline__ int marcel_polling_is_required(unsigned polling_point)
TBX_UNUSED;
static __tbx_inline__ void marcel_check_polling(unsigned polling_point)
TBX_UNUSED;
#section marcel_inline
void TBX_EXTERN __marcel_check_polling(unsigned polling_point);

static __tbx_inline__ int marcel_polling_is_required(unsigned polling_point)
{
	return !list_empty(&ma_ev_list_poll);
}

static __tbx_inline__ void marcel_check_polling(unsigned polling_point)
{
	if(marcel_polling_is_required(polling_point))
		__marcel_check_polling(polling_point);
}

/****************************************************************
 * Compatibility stuff
 *
 * Software using this should switch to the new interface
 * This will be remove later
 */

#section types

struct poll_struct;
struct poll_cell;

typedef struct poll_struct *marcel_pollid_t;
typedef struct poll_cell poll_cell_t, *marcel_pollinst_t;

#section marcel_structures
struct poll_struct {
	struct marcel_ev_server server;
	marcel_poll_func_t func;
	marcel_fastpoll_func_t fastfunc;
	marcel_pollgroup_func_t gfunc;
	void *specific;
	marcel_pollinst_t cur_cell;
};

struct poll_cell {
	struct marcel_ev_req inst;
	any_t arg;
};

#section types
typedef void (*marcel_pollgroup_func_t)(marcel_pollid_t id);

typedef void *(*marcel_poll_func_t)(marcel_pollid_t id,
				    unsigned active,
				    unsigned sleeping,
				    unsigned blocked);

typedef void *(*marcel_fastpoll_func_t)(marcel_pollid_t id,
					any_t arg,
					tbx_bool_t first_call);

#section macros
#define MARCEL_POLL_AT_TIMER_SIG  MARCEL_EV_POLL_AT_TIMER_SIG
#define MARCEL_POLL_AT_YIELD      MARCEL_EV_POLL_AT_YIELD
#define MARCEL_POLL_AT_LIB_ENTRY  MARCEL_EV_POLL_AT_LIB_ENTRY
#define MARCEL_POLL_AT_IDLE       MARCEL_EV_POLL_AT_IDLE

#section functions
marcel_pollid_t __tbx_deprecated__
marcel_pollid_create_X(marcel_pollgroup_func_t g,
		       marcel_poll_func_t f,
		       marcel_fastpoll_func_t h,
		       unsigned polling_points,
		       char* name);
#define marcel_pollid_create(g, f, h, polling_points) \
  marcel_pollid_create_X(g, f, h, polling_points, \
                         "Auto created ("__BASE_FILE__")")

#section macros
#define MARCEL_POLL_FAILED                NULL
#define MARCEL_POLL_OK                    (void*)1
#define MARCEL_POLL_SUCCESS(id)           ({MARCEL_EV_REQ_SUCCESS(&(id)->cur_cell->inst); (void*)1;})
#define MARCEL_POLL_SUCCESS_FOR(pollinst) ({MARCEL_EV_REQ_SUCCESS(&(pollinst)->inst); (void*)1;})


/*  ATTENTION : changement d interface */
/*  Remplacer : "FOREACH_POLL(id, _arg) { ..." */
/*  par : "FOREACH_POLL(id) { GET_ARG(id, _arg); ..." */
/*  Ou mieux: utiliser FOREACH_EV_POLL[_BASE] */
#define FOREACH_POLL(id) \
  FOREACH_REQ_POLL((id)->cur_cell, &(id)->server, inst)
#define GET_ARG(id, _arg) \
	_arg = (__typeof__(_arg))((id)->cur_cell->arg)

#define GET_CURRENT_POLLINST(id) ((id)->cur_cell)

#section functions
void __tbx_deprecated__ marcel_poll(marcel_pollid_t id, any_t arg);

void __tbx_deprecated__ marcel_force_check_polling(marcel_pollid_t id);

static __tbx_inline__ void __tbx_deprecated__ 
marcel_pollid_setspecific(marcel_pollid_t id, void *specific)
{
	id->specific = specific;
}

void marcel_poll_lock(void);
void marcel_poll_unlock(void);

static __tbx_inline__ void * __tbx_deprecated__ 
marcel_pollid_getspecific(marcel_pollid_t id)
{
	return id->specific;
}

